#include <vector>
#include <algorithm>
#include <set>

#include<vcg/simplex/vertex/base.h>
#include<vcg/simplex/face/base.h>
#include<vcg/simplex/face/component_ocf.h>
#include<vcg/complex/trimesh/base.h>

#include<vcg/complex/trimesh/update/flag.h>
#include<vcg/simplex/face/pos.h>
#include<vcg/simplex/face/topology.h>
#include<vcg/complex/trimesh/update/topology.h>
#include<vcg/complex/trimesh/update/normal.h>
#include<vcg/complex/trimesh/clean.h>

#include<wrap/io_trimesh/import_ply.h>
#include<wrap/io_trimesh/export_ply.h>

using namespace vcg;

class MyEdge;    // dummy prototype never used
class MyFace;
class MyVertex;

class MyVertex  : public VertexSimp2< MyVertex, MyEdge, MyFace, vertex::Coord3f, vertex::BitFlags, vertex::Normal3f, vertex::Color4b >{};
class MyFace    : public FaceSimp2  < MyVertex, MyEdge, MyFace, face::VertexRef, face::Normal3f, face::BitFlags, face::FFAdj, face::Color4b> {};
class MyMesh : public tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace > >{};

   void EdgeSwap(MyMesh::FacePointer f0, MyMesh::FacePointer f1,int e0, int e1)
    {
    //MyMesh::FacePointer f0,f1;
    //int  e0,e1; // lato condiviso

    assert(f0->FFp(e0)==f1);
    assert(f1->FFp(e1)==f0);

    MyMesh::VertexPointer ve0 = f0->V0(e0);
    MyMesh::VertexPointer ve1 = f0->V1(e0);

    assert( (ve0 == f1->V0(e1)) ||(ve0 == f1->V1(e1)) );
    assert( (ve1 == f1->V0(e1)) ||(ve1 == f1->V1(e1)) );

    assert(ve0 != ve1);

    MyMesh::VertexPointer vf0 = f0->V2(e0);
    MyMesh::VertexPointer vf1 = f1->V2(e1);

    assert(vf0 != vf1);

    f0->V0(e0) = vf1;
    f0->V1(e0) = vf0;
    f0->V2(e0) = ve0;

    f1->V0(e1) = vf0;
    f1->V1(e1) = vf1;
    f1->V2(e1) = ve1;

    MyMesh::FacePointer f00 = f0->FFp( (e0+1)%3 );   int e00 = f0->FFi( (e0+1)%3 );
    MyMesh::FacePointer f01 = f0->FFp( (e0+2)%3 );   int e01 = f0->FFi( (e0+2)%3 );
    MyMesh::FacePointer f10 = f1->FFp( (e1+1)%3 );   int e10 = f1->FFi( (e1+1)%3 );
    MyMesh::FacePointer f11 = f1->FFp( (e1+2)%3 );   int e11 = f1->FFi( (e1+2)%3 );

    if(! face::IsBorder(*f0,(e0+1)%3 ) ) {
        f0->FFp( (e0+1)%3 ) = f00;  f0->FFi( (e0+1)%3 ) = e00;
        f00->FFp(   e00    ) = f0;  f00->FFi(   e00    ) = (e0+1)%3 ;
    }
    if(! face::IsBorder(*f0,(e0+2)%3 ) ) {
         f0->FFp( (e0+2)%3 ) = f10;  f0->FFi( (e0+2)%3 ) = e10;
        f10->FFp(   e10    ) = f0;  f10->FFi(   e10    ) = (e0+2)%3 ;
    }
    if(! face::IsBorder(*f1,(e1+1)%3 ) ) {
         f1->FFp( (e1+1)%3 ) = f11;  f1->FFi( (e1+1)%3 ) = e11;
        f11->FFp(   e11    ) = f1;  f11->FFi(   e11    ) = (e1+1)%3 ;
    }
    if(! face::IsBorder(*f1,(e1+2)%3 ) ) {
         f1->FFp( (e1+2)%3 ) = f01;  f1->FFi( (e1+2)%3 ) = e01;
        f01->FFp(   e01    ) = f1;  f01->FFi(   e01    ) = (e1+2)%3 ;
    }
}

   bool SwapTest(MyMesh::FacePointer f0, MyMesh::FacePointer f1,int e0, int e1)
   {
       float thr = vcg::math::ToRad(5.0f);
       if(vcg::Angle(f0->N(),f1->N()) > thr) return false;

       float edgeLenOrig  = vcg::Distance(f0->V0(e0)->P(),f0->V1(e0)->P());
       float edgeLenAfter = vcg::Distance(f0->V2(e0)->P(),f1->V2(e1)->P());

       if(edgeLenOrig < edgeLenAfter) return false;

       return true;
   }
   int main(int argc,char **argv ){

    MyMesh m;
    if(argc==1) return 0;
   int errorCode = tri::io::ImporterPLY<MyMesh>::Open(m,argv[1]);
    if(errorCode!=0)
		{
      printf("Error reading file  %s :%s :\n",argv[1], tri::io::ImporterPLY<MyMesh>::ErrorMsg(errorCode) );
			exit(0);
		}
		
    printf("vn %i fn %i\n",m.vn, m.fn);

    MyMesh::VertexIterator vi;
    MyMesh::FaceIterator fi;
    vcg::tri::UpdateNormals<MyMesh>::PerFaceNormalized(m);
    vcg::tri::UpdateTopology<MyMesh>::FaceFace(m);
    vcg::tri::UpdateFlags<MyMesh>::FaceClearV(m);
    int swapCnt=0;

    for(fi=m.face.begin();fi!=m.face.end();++fi)
    {
        if(!(*fi).IsV())
        {
            for(int i=0;i<3;++i)
                if( (! face::IsBorder(*fi,i) ) &&  (! (*fi).FFp(i)->IsV()) )
                {
                    if(SwapTest(&*fi, (*fi).FFp(i), i, (*fi).FFi(i)) )
                    {
                        ++swapCnt;
                        for(int j=0;j<3;++j)
                        {
                            (*fi).FFp(j)->SetV();
                            (*fi).FFp(i)->FFp(j)->SetV();
                        }
                        EdgeSwap( &*fi, (*fi).FFp(i), i, (*fi).FFi(i) );
                        break;
                    }
                }
        }
    }
    assert(tri::Clean<MyMesh>::IsFFAdjacencyConsistent(m));
    printf("Swapped %i triangle pairs\n",swapCnt);
  vcg::tri::io::ExporterPLY<MyMesh>::Save(m,"test.ply");
  return 0;
}

